package com.danitheskunk.skunkworks.backends.gl;

import com.danitheskunk.skunkworks.*;
import com.danitheskunk.skunkworks.gfx.IPipeline;
import com.danitheskunk.skunkworks.gfx.ITexture;
import com.danitheskunk.skunkworks.gfx.Image;
import org.lwjgl.glfw.GLFWErrorCallback;
import org.lwjgl.opengl.GL;
import org.lwjgl.opengl.GL11;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static org.lwjgl.glfw.GLFW.*;
import static org.lwjgl.opengl.GL46.*;
import static org.lwjgl.system.MemoryUtil.NULL;

public class Window extends BaseWindow {

	private final Framebuffer framebuffer;
	private final Mouse mouse;
	private final Pipeline2D pipeline2D;
	private final Pipeline3D pipeline3D;
	private final Map<String, Filter> scalers;
	private final TextureAtlas textureAtlas;
	private final long window;
	private boolean shouldClose;
	private Vec2i windowSize;

	//private final FilterFXAA fxaa;

	public Window(Vec2i size, String title, Engine engine) {
		super(engine);
		GLFWErrorCallback.createPrint(System.err).set();
		if(!glfwInit()) throw new IllegalStateException(
			"Unable to initialize GLFW");

		glfwDefaultWindowHints();
		windowSize = size;


		window = glfwCreateWindow(size.getX(), size.getY(), title, NULL, NULL);
		if(window == NULL) throw new RuntimeException(
			"Failed to create GLFW window");


		glfwShowWindow(window);
		glfwMakeContextCurrent(window);
		glfwSwapInterval(1); //vsync
		GL.createCapabilities();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glEnable(GL_COLOR);
		glEnable(GL_TEXTURE_2D);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);
		glEnable(GL_DEPTH_TEST);

		glfwSetWindowSizeCallback(window, this::windowSizeCallback);

		framebuffer = new Framebuffer(size);
		textureAtlas = new TextureAtlas();

		pipeline2D = new Pipeline2D(framebuffer, textureAtlas);
		pipeline3D = new Pipeline3D(framebuffer, textureAtlas);

		scalers = new HashMap<>();
		scalers.put("stretch", new ScalerStretch(framebuffer));
		scalers.put("aspect", new ScalerAspect(framebuffer));
		scalers.put("integer", new ScalerInteger(framebuffer));
		scaler = "aspect";

		mouse = new Mouse(window, scalers.get(scaler));
		shouldClose = false;

		System.out.println("Skunkworks window initialised");
		System.out.print("Maximum Texture Size: ");
		System.out.println(GL11.glGetInteger(GL_MAX_TEXTURE_SIZE));
	}

	@Override
	public void finishFrame() {
		runScaler();
		glfwSwapBuffers(window);
	}

	@Override
	public BaseGamepad getGamepad(int id) {
		return new Gamepad(id);
	}

	@Override
	public IMouse getMouse() {
		return mouse;
	}


	@Override
	public IPipeline getPipeline2D() {
		return pipeline2D;
	}

	@Override
	public IPipeline getPipeline3D() {
		return pipeline3D;
	}

	@Override
	public ITexture loadTexture(Image image) {
		return textureAtlas.addTexture(image);
	}

	@Override
	public ITexture loadTexture(String path) {
		var img = engine.loadImage(path);
		return textureAtlas.addTexture(img);
	}

	@Override
	public List<ITexture> loadTextureArray(Image image, Vec2i tileSize) {
		var tileCount = Vec2i.div(image.getSize(), tileSize);
		var tiles = new ArrayList<ITexture>();

		for(int y = 0; y < tileCount.getY(); ++y) {
			for(int x = 0; x < tileCount.getX(); ++x) {
				var rect = new Recti(Vec2i.mul(new Vec2i(x, y), tileSize),
					tileSize
				);
				var img = image.getSubImage(rect);
				tiles.add(loadTexture(img));
			}
		}

		return tiles;
	}

	@Override
	public List<ITexture> loadTextureArray(String path, Vec2i tileSize) {
		var img = engine.loadImage(path);
		return loadTextureArray(img, tileSize);
	}


	@Override
	public void runScaler() {
		scalers.get(scaler).apply();
		scalers.get(scaler).getOutput().copyToScreen();
	}

	@Override
	public void setDebug(boolean on) {
		pipeline2D.setDebug(on);
	}

	@Override
	public void setRenderSize(Vec2i size) {
		framebuffer.setSize(size);
	}

	@Override
	public void setScaler(String scaler) {
		super.setScaler(scaler);
		mouse.setScaler(scalers.get(scaler));
	}

	@Override
	public void setWindowSize(Vec2i size) {
		glfwSetWindowSize(window, size.getX(), size.getY());
	}

	@Override
	public boolean shouldClose() {
		return shouldClose;
	}

	@Override
	public void startFrame() {
		framebuffer.bind();
		glClearColor(0.f, 0.f, 0.f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	}

	@Override
	public void tick() { //one tick per update call?
		glfwMakeContextCurrent(window);
		if(glfwWindowShouldClose(window)) {
			shouldClose = true;
			return;
		}

		mouse.tick();
		glfwPollEvents();
	}

	private void windowSizeCallback(
		@SuppressWarnings("unused") long window, int width, int height
	) {
		System.out.printf("new window size %d x %d\n", width, height);
		windowSize = new Vec2i(width, height);
		glViewport(0, 0, width, height);

		//todo: do we really wanna do it like this?
		for(var entry : scalers.entrySet()) {
			entry.getValue().setOutputSize(windowSize);
		}
	}
}
